import logging
import re
import os
import shutil

from tempfile import mkdtemp
from cached_property import cached_property
from munch import Munch

from . import utils

from .utils import run, QUIET, MODEST, VERBOSE, macro_helper_cmd
from .errors import RpkgError

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

log = logging.getLogger('__main__')

class GitClient(object):

    def __init__(self, pkg_path, clone_url=None, anon_clone_url=None,
                 default_namespace=None, push_follow_tags=True,
                 clean_force=True, clean_dirs=True, log_graph=True,
                 log_all=True, gpg_keyid=None, git_props=None,
                 verbosity=MODEST):

        self.pkg_path = pkg_path
        self.clone_url_template = clone_url
        self.anon_clone_url_template = anon_clone_url
        self.default_namespace = default_namespace
        self.push_follow_tags = push_follow_tags
        self.clean_force = clean_force
        self.clean_dirs = clean_dirs
        self.log_graph = log_graph
        self.log_all = log_all
        self.gpg_keyid = gpg_keyid
        self.git_props = git_props or GitClient.git_props(pkg_path)
        self.verbosity = verbosity
        self.gitignore_path = os.path.join(self.pkg_path, '.gitignore')

    ######################## helpers ########################

    @classmethod
    def git_props(cls, pkg_path):
        git_props = Munch()

        git_props.root = run(
            macro_helper_cmd('git_root'),
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.branch = run(
            macro_helper_cmd('git_branch'),
            env={'GIT_ROOT': git_props.root},
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.remote = run(
            macro_helper_cmd('git_remote'),
            env={'GIT_ROOT': git_props.root, 'GIT_BRANCH': git_props.branch},
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.remote_url = run(
            macro_helper_cmd('git_remote_url'),
            env={'GIT_ROOT': git_props.root, 'GIT_REMOTE': git_props.remote},
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.head = run(
            macro_helper_cmd('git_head'),
            env={'GIT_ROOT': git_props.root},
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.head_short = run(
            macro_helper_cmd('git_head_short'),
            env={'GIT_ROOT': git_props.root, 'GIT_HEAD': git_props.head},
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.status = run(
            macro_helper_cmd('git_status'),
            env={'GIT_ROOT': git_props.root},
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.merged_tag_refs = run(
            macro_helper_cmd('git_merged_tag_refs'),
            env={'GIT_ROOT': git_props.root, 'GIT_HEAD': git_props.head},
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.submodule_refs = run(
            macro_helper_cmd('git_submodule_refs'),
            env={'GIT_ROOT': git_props.root, 'GIT_HEAD': git_props.head},
            capture_stderr=True, throw=False,
            cwd=pkg_path).out

        git_props.remote_netloc = urlparse(git_props.remote_url).netloc
        git_props.dirty = 'yes' if (git_props.status or (
            git_props.root and not git_props.head)) else ''

        return git_props

    def write_ignores(self, filenames, append=True):
        ignored_filenames = []

        if append:
            try:
                ignore_file = open(self.gitignore_path, 'r')
                ignored_filenames = [os.path.basename(l.strip())
                                     for l in ignore_file.readlines()]
                ignore_file.close()
            except OSError:
                pass

            ignore_file = open(self.gitignore_path, 'a')
        else:
            ignore_file = open(self.gitignore_path, 'w')

        for filename in filenames:
            if not filename in ignored_filenames:
                ignore_file.write('/'+filename+'\n')

        ignore_file.close()

    def add_to_index(self, *args):
        run(['git', 'add'] + list(args))

    def show_diff_index(self):
        run(['git', 'diff', '--cached'], capture=False)

    def ls_tree(self, *args):
        return run(['git', 'ls-tree'] + list(args)).out

    def status(self, *args):
        return run(['git', 'status'] + list(args)).out

    def commit_id_from_tag_id(self, tag_id):
        return run(['git', 'rev-parse', tag_id+'^{commit}']).out

    def read_file(self, revision, file_path):
        file_path = os.path.realpath(file_path)
        git_root_path = os.path.realpath(self.git_props.root)

        tree_path = file_path.replace(git_root_path, '', 1).strip('/')
        cmd = ['git', 'show', revision+':'+tree_path]
        return run(cmd, capture_stderr=True).out

    def clone_url(self, anonymous, repo_path):
        if utils.has_scheme(repo_path):
            return repo_path

        repo_path = repo_path.strip('/')
        if len(repo_path.split('/')) == 1 and self.default_namespace:
            repo_path = os.path.join(self.default_namespace, repo_path)

        if anonymous:
            return self.anon_clone_url_template % {'repo_path': repo_path}
        else:
            return self.clone_url_template % {'repo_path': repo_path}

    def latest_merged_tag_id(self, package_name, branched=''):
        tag_refs = self.git_props.merged_tag_refs.splitlines()

        for tag_ref in tag_refs:
            tag_id, tagname = tag_ref.split()

            expr = r'^{0}-[^-]+-{1}[^-]+$'.format(package_name, branched)
            if re.match(expr, tagname):
                return tag_id

        return None

    def tag_message(self, new_tagname, skip_edit=False):
        tag_parts = new_tagname.rsplit('-', 2)
        package_name = tag_parts[0]
        release = tag_parts[2]

        if release.startswith(self.git_props.branch+'.'):
            latest_merged_tag_id = self.latest_merged_tag_id(package_name,
                                                             self.git_props.branch+'.')
        else:
            latest_merged_tag_id = self.latest_merged_tag_id(package_name)

        if latest_merged_tag_id:
            tagged_commit_id = self.commit_id_from_tag_id(latest_merged_tag_id)
            rev_list_expr = '{0}..{1}'.format(tagged_commit_id, self.git_props.head)
        else:
            rev_list_expr = self.git_props.head

        log_cmd = ['git', 'log', '--format=%s', rev_list_expr, self.pkg_path]
        commit_messages = run(log_cmd).out.split('\n')
        entries = [utils.changelog_entry(msg) for msg in commit_messages]

        editor = run(['git', 'var', 'GIT_EDITOR']).out
        subject = 'Commits: {0}'.format(rev_list_expr)

        help_text = ("#\n"
        "# Write a changelog for tag:\n"
        "#   " + new_tagname + "\n"
        "# Lines starting with '#' will be ignored.\n")

        body = '\n'.join(entries) + '\n' + help_text
        if not skip_edit:
            body = utils.edit(editor, body)

        return '{0}\n\n{1}'.format(subject, body)

    def repo_path(self):
        remote_url = self.git_props.remote_url

        replacements = {'repo_path': '(.*)/?'}
        clone_url_pattern = '^' + self.clone_url_template % replacements + '$'
        anon_clone_url_pattern = '^' + self.anon_clone_url_template % replacements + '$'

        match = re.match(clone_url_pattern, remote_url)
        if not match:
            match = re.match(anon_clone_url_pattern, remote_url)

        if match and len(match.groups()) > 0:
            repo_path = match.group(1)
        else:
            repo_path = urlparse(remote_url).path.rsplit(':')[-1].strip('/')

        if repo_path.endswith('.git'):
            repo_path = repo_path[:-len('.git')]

        return repo_path

    ######################## commands ########################

    def add(self, all=False, paths=[]):
        cmd = ['git', 'add']

        if all:
            cmd.extend(['--all'])

        if paths:
            cmd.extend(['--'])

        cmd.extend(paths)
        run(cmd, capture=False)

    def switch(self, target, create=False, detach=False):
        cmd = ['git', 'switch']

        if create:
            cmd.extend(['-c'])

        if detach:
            cmd.extend(['-d'])

        cmd.extend([target])
        run(cmd, capture=False)

    def clean(self, dry_run=False, x=False, force=False, d=False):
        cmd = ['git', 'clean']

        if dry_run:
            cmd.extend(['-n'])

        if x:
            cmd.extend(['-x'])

        if force or self.clean_force:
            cmd.extend(['-f'])

        if d or self.clean_dirs:
            cmd.extend(['-d'])

        if self.verbosity == QUIET:
            cmd.extend(['-q'])

        run(cmd, capture=False)

    def clone(self, repo_path, directory=None, branch=None, anonymous=False):
        if not directory:
            directory = os.path.basename(repo_path)

        cmd = ['git', 'clone',
               self.clone_url(anonymous, repo_path),
               directory]

        if branch:
            cmd.extend(['--branch', branch])

        run(cmd, capture=False)

    def clone_with_dirs(self, repo_path, directory, anonymous=False):
        os.makedirs(directory, exist_ok=True)
        bare_dir = mkdtemp(dir=directory)
        clone_url = self.clone_url(anonymous, repo_path)

        clone_cmd = ['git', 'clone', '--bare', clone_url, bare_dir]
        run(clone_cmd, capture=False)

        branches_cmd = ['git', 'branch', '--format', '%(refname:short)']
        branches = run(branches_cmd, cwd=bare_dir).out.split()

        for branch in branches:
            branch_dir = os.path.join(directory, branch)
            run(['git', 'clone', '--branch', branch, bare_dir, branch_dir])
            run(['git', 'remote', 'set-url', 'origin', clone_url],
                cwd=branch_dir)

        shutil.rmtree(bare_dir)

    def commit(self, all=False, amend=False, msg_lines=[],
               msg_file=None, sign_off=False, files=[]):
        cmd = ['git', 'commit']

        if self.verbosity == QUIET:
            cmd.extend(['-q'])

        if all:
            cmd.extend(['-a'])

        if amend:
            cmd.extend(['--amend'])

        for msg_line in msg_lines:
            cmd.extend(['-m', msg_line])

        if msg_file:
            cmd.extend(['-F', msg_file])

        if sign_off:
            cmd.extend(['-s'])

        if self.gpg_keyid:
            cmd.extend(['-S'+self.gpg_keyid])

        if files:
            cmd.extend(['--'])

        cmd.extend(files)
        run(cmd, capture=False)

    def diff(self, cached=False, files=[]):
        cmd = ['git', 'diff']

        if cached:
            cmd.extend(['--cached'])

        if files:
            cmd.extend(['--'])

        cmd.extend(files)
        run(cmd, capture=False)

    def fetch(self):
        cmd = ['git', 'fetch']
        run(cmd, capture=False)

    def log(self, graph=False, all=False,
            show_signature=False, positionals=[]):
        cmd = ['git', 'log']

        if graph or self.log_graph:
            cmd.extend(['--graph'])

        if all or self.log_all:
            cmd.extend(['--all'])

        if show_signature:
            cmd.extend(['--show-signature'])

        cmd.extend(positionals)
        run(cmd, capture=False)

    def list_branches(self):
        cmd = ['git', 'branch', '-a']
        run(cmd, capture=False)

    def merge_continue(self):
        cmd = ['git', 'merge', '--continue']
        run(cmd, capture=False)

    def merge_abort(self):
        cmd = ['git', 'merge', '--abort']
        run(cmd, capture=False)

    def merge(self, no_commit=False, squash=False,
              message=None, msg_file=None,
              sign_off=False, commits=[]):
        cmd = ['git', 'merge']

        if no_commit:
            cmd.extend(['--no-commit'])

        if squash:
            cmd.extend(['--squash'])

        if message:
            cmd.extend(['-m', message])

        if msg_file:
            cmd.extend(['-F', msg_file])

        if sign_off:
            cmd.extend(['--signoff'])

        if self.gpg_keyid:
            cmd.extend(['-S'+self.gpg_keyid])

        if commits:
            cmd.extend(['--'])

        cmd.extend(commits)
        run(cmd, capture=False)

    def new(self, package_name):
        tag_id = self.latest_merged_tag_id(package_name)
        run(['git', 'diff', tag_id], capture=False)

    def pull(self, rebase=False, no_rebase=False):
        cmd = ['git', 'pull']

        if self.verbosity == QUIET:
            cmd.extend(['-q'])

        if rebase:
            cmd.extend(['--rebase'])

        if no_rebase:
            cmd.extend(['--no-rebase'])

        run(cmd, capture=False)

    def push(self, force=False, follow_tags=False):
        cmd = ['git', 'push']

        if force:
            cmd.extend(['-f'])

        if follow_tags or self.push_follow_tags:
            cmd.extend(['--follow-tags'])

        if self.verbosity == QUIET:
            cmd.extend(['-q'])

        run(cmd, capture=False)

    def reset(self, mode=0, positional=None):
        cmd = ['git', 'reset']

        if mode == 1:
            cmd.extend(['--mixed'])
        elif mode == 2:
            cmd.extend(['--soft'])
        elif mode == 3:
            cmd.extend(['--hard'])

        if positional:
            cmd.extend([positional])

        run(cmd, capture=False)

    def status(self, pathspecs=[]):
        cmd = ['git', 'status']

        if pathspecs:
            cmd.extend(['--'])

        cmd.extend(pathspecs)
        run(cmd, capture=False)

    def add_tag(self, tagname, skip_edit=False):
        """Add a git tag to the repository

        Takes an optional tagname.
        Runs the tag command and returns nothing.
        """
        if not re.match(r'^.*-[^-]+-[^-]+$', tagname):
            raise RpkgError("Tag {0} does not have n-v-r format.".format(tagname))

        if tagname in run(['git', 'tag']).out:
            raise RpkgError("Tag {0} already exists.".format(tagname))

        if self.gpg_keyid:
            cmd = ['git', 'tag', '-u', self.gpg_keyid]
        else:
            cmd = ['git', 'tag', '-a']

        cmd.extend(['-m', self.tag_message(tagname, skip_edit), tagname])

        run(cmd)
        log.info('Created tag %s', tagname)

    def list_tags(self, tagname=None):
        cmd = ['git', 'tag', '-l']

        if tagname:
            cmd.extend([tagname])

        run(cmd, capture=False)

    def delete_tag(self, tagname=None):
        cmd = ['git', 'tag', '-d', tagname]
        run(cmd, capture=False)
